@@ -1,11 +1,15 @@
-
-from copy import copy
+from contextlib import contextmanager
 
 try:
     from django.forms.utils import flatatt
-except ImportError:  # Django 1.5 compatibility
-    from django.forms.util import flatatt
+except ImportError:  # django 1.4
+    from django.forms.util import flatatt as flatatt_
+    from django.utils.safestring import mark_safe
+    def flatatt(text):
+        return mark_safe(flatatt_(text))
+
 from django import template
+from django.db.models.fields.files import FieldFile, ImageFile
 from django.template.base import token_kwargs
 from django.template.loader import get_template
 from django.template.loader_tags import (
@@ -35,7 +39,7 @@ def resolve_blocks(template, context):
     '''
     Return a BlockContext instance of all the {% block %} tags in the template.
 
-    If template is a string, it will be resovled through get_template
+    If template is a string, it will be resolved through get_template
     '''
     try:
         blocks = context.render_context[BLOCK_CONTEXT_KEY]
@@ -46,6 +50,9 @@ def resolve_blocks(template, context):
     if isinstance(template, six.string_types):
         template = get_template(template)
 
+    # For Django 1.8 compatibility
+    template = getattr(template, 'template', template)
+
     # Add this templates blocks as the first
     local_blocks = dict(
         (block.name, block)
@@ -80,31 +87,49 @@ def parse_widget_name(widget):
     return alias, block_name
 
 
-def lookup_block(context, alias, *names):
+@contextmanager
+def using(context, alias):
     '''
-    Find the first available block in a given alias set.
-
-    Samurai.
+    Temporarily update the context to use the BlockContext for the given alias.
     '''
 
-    try:
-        widgets = context.render_context[WIDGET_CONTEXT_KEY]
-    except KeyError:
-        raise template.TemplateSyntaxError("No widget libraries loaded!")
+    # An empty alias means look in the current widget set.
+    if alias == '':
+        yield context
+    else:
+        try:
+            widgets = context.render_context[WIDGET_CONTEXT_KEY]
+        except KeyError:
+            raise template.TemplateSyntaxError('No widget libraries loaded!')
 
-    try:
-        block_set = widgets[alias]
-    except KeyError:
-        raise template.TemplateSyntaxError(
-            'No widget library loaded for alias: %r' % alias
-        )
+        try:
+            block_set = widgets[alias]
+        except KeyError:
+            raise template.TemplateSyntaxError(
+                'No widget library loaded for alias: %r' % alias
+            )
 
+        context.render_context.push()
+        context.render_context[BLOCK_CONTEXT_KEY] = block_set
+        context.render_context[WIDGET_CONTEXT_KEY] = widgets
+
+        yield context
+
+        context.render_context.pop()
+
+
+def find_block(context, *names):
+    '''
+    Find the first matching block in the current block_context
+    '''
+    block_set = context.render_context[BLOCK_CONTEXT_KEY]
     for name in names:
         block = block_set.get_block(name)
         if block is not None:
             return block
+
     raise template.TemplateSyntaxError(
-        'No widget found in %r for: %r' % (alias, names)
+        'No widget found for: %r' % (names,)
     )
 
 
@@ -113,6 +138,7 @@ def load_widgets(context, **kwargs):
     '''
     Load a series of widget libraries.
     '''
+    _soft = kwargs.pop('_soft', False)
 
     try:
         widgets = context.render_context[WIDGET_CONTEXT_KEY]
@@ -120,57 +146,117 @@ def load_widgets(context, **kwargs):
         widgets = context.render_context[WIDGET_CONTEXT_KEY] = {}
 
     for alias, template_name in kwargs.items():
-        # Build an isolated render context each time
-        safe_context = copy(context)
-        safe_context.render_context = safe_context.render_context.new({
-            BLOCK_CONTEXT_KEY: BlockContext(),
-        })
-        blocks = resolve_blocks(template_name, safe_context)
-        widgets[alias] = blocks
+        if _soft and alias in widgets:
+            continue
+
+        context.render_context.push()
+        context.render_context[BLOCK_CONTEXT_KEY] = BlockContext()
+
+        try:
+            blocks = resolve_blocks(template_name, context)
+            widgets[alias] = blocks
+        finally:
+            context.render_context.pop()
 
     return ''
 
 
-@register.simple_tag(takes_context=True)
-def widget(context, widget, **kwargs):
-    alias, block_name = parse_widget_name(widget)
-
-    block = lookup_block(context, alias, block_name)
-
-    context.update(kwargs)
-    try:
-        return block.render(context)
-    finally:
-        context.pop()
+def pop_asvar(bits):
+    if len(bits) >= 2 and bits[-2] == 'as':
+        asvar = bits[-1]
+        del bits[-2:]
+        return asvar
 
 
-class NestedWidget(template.Node):
-    def __init__(self, widget, nodelist, kwargs):
+class Widget(template.Node):
+    def __init__(self, widget, kwargs, asvar):
         self.widget = widget
-        self.nodelist = nodelist
         self.kwargs = kwargs
+        self.asvar = asvar
 
     def render(self, context):
         widget = self.widget.resolve(context)
 
         alias, block_name = parse_widget_name(widget)
 
-        block = lookup_block(context, alias, block_name)
+        with using(context, alias):
+            block = find_block(context, block_name)
 
-        kwargs = {
-            key: val.resolve(context)
-            for key, val in self.kwargs.items()
-        }
-        context.update(kwargs)
-        try:
-            content = self.nodelist.render(context)
+            kwargs = {
+                key: val.resolve(context)
+                for key, val in self.kwargs.items()
+            }
+            context.update(kwargs)
             try:
-                context.update({'content': content})
-                return block.render(context)
+                result = block.render(context)
             finally:
                 context.pop()
-        finally:
-            context.pop()
+
+            if self.asvar:
+                context[self.asvar] = result
+                return ''
+
+            return result
+
+
+@register.tag
+def widget(parser, token):
+    bits = token.split_contents()
+    tag_name = bits.pop(0)
+
+    try:
+        widget = parser.compile_filter(bits.pop(0))
+    except IndexError:
+        raise template.TemplateSyntaxError(
+            '%s requires one positional argument' % tag_name
+        )
+
+    asvar = pop_asvar(bits)
+
+    kwargs = token_kwargs(bits, parser)
+    if bits:
+        raise template.TemplateSyntaxError(
+            '%s accepts only one positional argument' % tag_name
+        )
+
+    return Widget(widget, kwargs, asvar)
+
+
+class NestedWidget(template.Node):
+    def __init__(self, widget, nodelist, kwargs, asvar):
+        self.widget = widget
+        self.nodelist = nodelist
+        self.kwargs = kwargs
+        self.asvar = asvar
+
+    def render(self, context):
+        widget = self.widget.resolve(context)
+
+        alias, block_name = parse_widget_name(widget)
+
+        with using(context, alias):
+            block = find_block(context, block_name)
+
+            kwargs = {
+                key: val.resolve(context)
+                for key, val in self.kwargs.items()
+            }
+            context.update(kwargs)
+            try:
+                content = self.nodelist.render(context)
+                try:
+                    context.update({'content': content})
+                    result = block.render(context)
+                finally:
+                    context.pop()
+            finally:
+                context.pop()
+
+            if self.asvar:
+                context[self.asvar] = result
+                return ''
+
+            return result
 
 
 @register.tag
@@ -185,7 +271,10 @@ def nested_widget(parser, token):
             '%s requires one positional argument' % tag_name
         )
 
+    asvar = pop_asvar(bits)
+
     kwargs = token_kwargs(bits, parser)
+
     if bits:
         raise template.TemplateSyntaxError(
             '%s accepts only one positional argument' % tag_name
@@ -194,7 +283,7 @@ def nested_widget(parser, token):
     nodelist = parser.parse(('endnested',))
     parser.delete_first_token()
 
-    return NestedWidget(widget, nodelist, kwargs)
+    return NestedWidget(widget, nodelist, kwargs, asvar)
 
 
 @register.simple_tag(takes_context=True)
@@ -202,40 +291,76 @@ def form_field(context, field, widget=None, **kwargs):
     if widget is None:
         alias = kwargs.pop('alias', 'form')
 
-        block = lookup_block(context, alias, *auto_widget(field))
+        block_names = auto_widget(field)
     else:
         alias, block_name = parse_widget_name(widget)
 
-        block = lookup_block(context, alias, block_name)
+        block_names = [block_name]
 
     field_data = {
         'form_field': field,
         'id': field.auto_id,
+        'widget_type': field.field.widget.__class__.__name__,
+        'field_type': field.field.__class__.__name__,
     }
 
     for attr in ('css_classes', 'errors', 'field', 'form', 'help_text',
-                 'html_name', 'id_for_label', 'label', 'name', 'value',):
+                 'html_name', 'id_for_label', 'label', 'name',):
         field_data[attr] = getattr(field, attr)
 
     for attr in ('choices', 'widget', 'required'):
         field_data[attr] = getattr(field.field, attr, None)
 
+    # Grab the calculated value
+    value = field.value()
+
+    # If we have choices, help out some
     if field_data['choices']:
+        if isinstance(value, (list, tuple)):
+            # XXX Is there any value in providing a separate display list?
+            pass
+        else:
+            field_data['display'] = dict(field.field.choices).get(value, '')
+
         field_data['choices'] = [
             (force_text(k), v)
             for k, v in field_data['choices']
         ]
-        # Normalize the value [django.forms.widgets.Select.render_options]
-        field_data['value'] = force_text(field_data['value']())
+
+    # If it's a FileField, expose attributes of FieldFile that might be useful
+    if value and isinstance(value, FieldFile):
+        file_data = {
+            'size': value.size,
+            'url': value.url,
+        }
+        if isinstance(value, ImageFile):
+            file_data.update({
+                'width': value.width,
+                'height': value.height,
+            })
+
+        field_data['file'] = file_data
+
+    # Normalize the value [django.forms.widgets.Select.render_options]
+    if value is None:
+        pass
+    elif isinstance(value, (list, tuple)):
+        value = tuple(map(force_text, value))
+    else:
+        value = force_text(value)
+    field_data['value'] = value
 
     # Allow supplied values to override field data
     field_data.update(kwargs)
 
-    context.update(field_data)
-    try:
-        return block.render(context)
-    finally:
-        context.pop()
+    with using(context, alias):
+        block = find_block(context, *block_names)
+
+        context.update(field_data)
+        try:
+            return block.render(context)
+        finally:
+            context.pop()
 
 
 def auto_widget(field):
@@ -272,19 +397,24 @@ def reuse(context, block_list, **kwargs):
     Allow reuse of a block within a template.
 
     {% reuse '_myblock' foo=bar %}
+
+    If passed a list of block names, will use the first that matches:
+
     {% reuse list_of_block_names .... %}
     '''
-    block_context = context.render_context[BLOCK_CONTEXT_KEY]
+    try:
+        block_context = context.render_context[BLOCK_CONTEXT_KEY]
+    except KeyError:
+        block_context = BlockContext()
 
     if not isinstance(block_list, (list, tuple)):
         block_list = [block_list]
 
-    for name in block_list:
-        block = block_context.get_block(name)
-        if block is not None:
+    for block in block_list:
+        block = block_context.get_block(block)
+        if block:
             break
-
-    if block is None:
+    else:
         return ''
 
     context.update(kwargs)
